'''Result class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''
from .keyword import Keyword
from .token import Token
from .tokens import Tokens
from .endofstatement import _EndOfStatement


TOSTR = (Keyword, Token, Tokens, _EndOfStatement)


class Result:

    __slots__ = ('is_valid', 'pos', 'expecting', 'tree')

    def __init__(self, is_valid, pos):
        self.is_valid = is_valid
        self.pos = pos
        self.expecting = None
        self.tree = None

    def as_str(self, translate=None, line_number=False):
        if self.is_valid:
            return 'parsed successfully'
        if line_number:
            string = self.tree._string
            lnr = string.count('\n', 0, self.pos) + 1
            pos = self.pos - string.rfind('\n', 0, self.pos)
            res = ['error at line {}, col {}'.format(lnr, pos)]
        else:
            res = ['error at position {}'.format(self.pos)]
        arr = []
        for elem in (self.expecting):
            expectstr = translate(elem) if translate else None
            if expectstr is None and isinstance(elem, TOSTR):
                expectstr = str(elem)

            if expectstr:
                arr.append(expectstr)

        arr.sort()
        total = len(arr) - 1
        for i, expectstr in enumerate(arr):
            if i == 0:
                res.append(', expecting: {}'.format(expectstr))
            elif i < total:
                res.append(', {}'.format(expectstr))
            else:
                res.append(' or {}'.format(expectstr))

        return ''.join(res)

    def __str__(self):
        return self.as_str()



'''Token class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''
from .elements import NamedElement, c_export, go_export, java_export


class Token(NamedElement):

    __slots__ = ('_token',)

    def __init__(self, token):
        if not isinstance(token, str):
            raise TypeError('Token(): first argument must be a string '
                            'value, got {}'.format(token))
        self._token = token

    def __repr__(self):
        return self._token

    def _get_node_result(self, root, tree, rule, s, node):
        is_valid = s.startswith(self._token)

        if is_valid:
            root._append_tree(tree, node, node.start + len(self._token))
        else:
            root._expecting.update(self, node.start)

        return is_valid, node.end or node.start

    def _run_export_js(self, js_indent, indent, classes, cname):
        return 'Token(\'{}\')'.format(self._token.replace('\'', '\\\''))

    def _run_export_py(self, py_indent, indent, classes):
        return 'Token(\'{}\')'.format(self._token.replace('\'', '\\\''))

    @c_export
    def _run_export_c(self, js_indent, indent, enums, gid):
        return 'cleri_token({}, "{}")'.format(
            gid,
            self._token.replace('"', '\\"'))

    @go_export
    def _run_export_go(self, go_indent, indent, enums, gid):
        return 'goleri.NewToken({}, "{}")'.format(
            gid,
            self._token.replace('"', '\\"'))

    @java_export
    def _run_export_java(self, java_indent, indent, enums, classes, gid):
        return 'new Token({}"{}")'.format(
            '' if gid is None else 'Ids.{}, '.format(gid),
            self._token.replace('"', '\\"'))



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Keyword,
    Optional,
)  # nopep8


class TestOptinal(unittest.TestCase):

    def test_optional(self):
        k_hi = Keyword('hi')
        optional = Optional(k_hi)
        grammar = create_grammar(optional)

        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('hello').is_valid)

        self.assertEqual(
            str(grammar.parse('hello')),
            'error at position 0, expecting: end_of_statement or hi'
        )
        self.assertEqual(
            str(grammar.parse('hi hi')),
            'error at position 2, expecting: end_of_statement'
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Regex,
)  # nopep8


class TestRegex(unittest.TestCase):

    def test_regex(self):
        regex = Regex('^(?:\'(?:[^\']*)\')+')
        grammar = create_grammar(regex)

        self.assertTrue(grammar.parse("'hi'").is_valid)
        self.assertTrue(grammar.parse("'this is ''SiriDB'''").is_valid)
        self.assertFalse(grammar.parse("'Hi !").is_valid)
        self.assertFalse(grammar.parse("'hello''").is_valid)
        self.assertFalse(grammar.parse("").is_valid)
        self.assertEqual(
            grammar.parse('hi.').as_str(
                translate=lambda elem: 'single_quoted_string'
                if elem is regex else None),
            'error at position 0, expecting: single_quoted_string'
        )
        self.assertTrue(grammar.parse("'hi'").is_valid)

    def test_flags_regex(self):
        regex = Regex(r'(?s)//.*?(\r?\n|$)')
        grammar = create_grammar(regex)

        self.assertTrue(grammar.parse("""
                //test
        """).is_valid)


if __name__ == '__main__':
    unittest.main()



'''Expecting Class.

Expecting Class is used to return possible elements at a position when a given
statement is not correct. This helps building auto-completion, suggestions or
returning nice messages to the user.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''


class Expecting:

    __slots__ = ('required', 'optional', 'pos', '_modes')

    def __init__(self):
        self.required = set()
        self.optional = set()
        self.pos = 0
        self._modes = {self.pos: self.required}

    def set_mode_required(self, pos, is_required):
        # do nothing when mode is already set to optional
        if pos in self._modes and self._modes[pos] is self.optional:
            return

        self._modes[pos] = self.required if is_required else self.optional

    def empty(self):
        self.required.clear()
        self.optional.clear()

    def update(self, element, pos):
        if pos > self.pos:
            self.empty()
            self.pos = pos

        if pos == self.pos:
            self._modes[pos].add(element)

    def get_expecting(self):
        return self.required | self.optional



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Sequence,
    Keyword,
)  # nopep8


class TestSequence(unittest.TestCase):

    def test_sequence(self):
        k_hi = Keyword('hi')
        k_iris = Keyword('iris')
        seq = Sequence(k_hi, k_iris)
        grammar = create_grammar(seq)

        self.assertTrue(grammar.parse('hi iris').is_valid)
        self.assertFalse(grammar.parse(' hi sasha ').is_valid)
        self.assertEqual(
            str(grammar.parse('hi sasha')),
            'error at position 3, expecting: iris'
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Sequence,
    Choice,
    Keyword,
)  # nopep8


class TestChoice(unittest.TestCase):

    def test_choice_most_greedy(self):
        k_hi = Keyword('hi')
        k_iris = Keyword('iris')
        seq = Sequence(k_hi, k_iris)
        choice = Choice(k_hi, seq)
        grammar = create_grammar(choice)

        self.assertTrue(choice.most_greedy)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse(' hi iris ').is_valid)
        self.assertFalse(grammar.parse(' hi sasha ').is_valid)

    def test_choice_first_match(self):
        k_hi = Keyword('hi')
        k_iris = Keyword('iris')
        seq = Sequence(k_hi, k_iris)
        choice = Choice(k_hi, seq, most_greedy=False)
        grammar = create_grammar(choice)

        self.assertFalse(choice.most_greedy)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertFalse(grammar.parse(' hi iris ').is_valid)
        self.assertFalse(grammar.parse(' hi sasha ').is_valid)


if __name__ == '__main__':
    unittest.main()



'''Node class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''


class Node:

    __slots__ = ('element', '_string', 'start', 'end', 'children')

    def __init__(self, element, string, start, end=None):
        self.element = element
        self.start = start
        self.end = end
        self._string = string
        self.children = []

    @property
    def string(self):
        return self._string[self.start:self.end]



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Keyword,
    Token,
    Choice,
)  # nopep8


class TestKeyword(unittest.TestCase):

    def test_keyword(self):
        k_hi = Keyword('hi')
        grammar = create_grammar(k_hi)

        self.assertFalse(k_hi.ign_case)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse(' hi ').is_valid)
        self.assertFalse(grammar.parse('Hi').is_valid)
        self.assertFalse(grammar.parse('hello').is_valid)
        self.assertEqual(
            'error at position 0, expecting: hi',
            str(grammar.parse('hello')))
        self.assertEqual(
            'parsed successfully',
            str(grammar.parse('hi')))

    def test_keyword_ign_case(self):
        k_hi = Keyword('hi', ign_case=True)
        grammar = create_grammar(k_hi)

        self.assertTrue(k_hi.ign_case)
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse('Hi').is_valid)
        self.assertFalse(grammar.parse('hello').is_valid)
        self.assertEqual(
            'error at position 2, expecting: end_of_statement',
            str(grammar.parse('Hi Iris')))

    def test_keyword_alt_regkw(self):
        grammar = create_grammar(Choice(Keyword('hi'), Token('HI')), r'[a-z]+')
        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse('HI').is_valid)
        self.assertFalse(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('Hi').is_valid)


if __name__ == '__main__':
    unittest.main()



'''This class and THIS instance.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''
from .elements import Element


class This(Element):

    def _get_node_result(self, root, tree, rule, _s, node):
        if node.start not in rule._tested:
            rule._tested[node.start] = root._walk(
                rule._element,
                node.start,
                node.children,
                rule,
                True)
            rule._tree[node.start] = node.children

        if rule._tested[node.start][0]:
            node.children = rule._tree[node.start]
            root._append_tree(tree, node, rule._tested[node.start][1])

        return rule._tested[node.start]

    def _export_js(self, js_indent, indent, classes, cname):
        classes.add('THIS')
        return 'THIS'

    def _export_py(self, py_indent, indent, classes):
        classes.add('THIS')
        return 'THIS'

    def _export_c(self, c_indent, indent, enums):
        return 'CLERI_THIS'

    def _export_go(self, go_indent, indent, enums):
        return 'goleri.THIS'

    def _export_java(self, java_indent, indent, enums, classes):
        classes.add('jleri.This')
        return 'This.THIS'


THIS = This()



import unittest
import os
import sys
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    KeywordError,
    create_grammar,
    Sequence,
    Keyword,
    Prio,
    List,
    THIS,
    MaxRecursionError,
)  # nopep8


class TestPrio(unittest.TestCase):

    def test_prio(self):
        k_hi = Keyword('hi')
        k_bye = Keyword('bye')
        k_and = Keyword('and')
        k_or = Keyword('or')
        prio = Prio(
            k_hi,
            k_bye,
            Sequence('(', THIS, ')'),
            Sequence(THIS, k_and, THIS),
            Sequence(THIS, k_or, THIS),
        )
        grammar = create_grammar(prio)

        self.assertTrue(grammar.parse('hi').is_valid)
        self.assertTrue(grammar.parse('(bye)').is_valid)
        self.assertTrue(grammar.parse('(hi and bye)').is_valid)
        self.assertTrue(grammar.parse('(hi or hi) and (hi or hi)').is_valid)
        self.assertTrue(grammar.parse('(hi or (hi and bye))').is_valid)
        self.assertFalse(grammar.parse('').is_valid)
        self.assertFalse(grammar.parse('(hi').is_valid)
        self.assertFalse(grammar.parse('()').is_valid)
        self.assertFalse(grammar.parse('(hi or hi) and').is_valid)
        with self.assertRaises(MaxRecursionError):
            grammar.parse(
                '(((((((((((((((((((((((((((((((((((((((((((((((((((hi'
                ')))))))))))))))))))))))))))))))))))))))))))))))))))'
            )

    def test_prio_list(self):
        arr = Sequence('[', List(THIS), ']')
        prio = Prio(Keyword('nil'), arr)
        grammar = create_grammar(prio)
        self.assertTrue(grammar.parse('[]').is_valid)
        self.assertTrue(grammar.parse('[[nil]]').is_valid)
        self.assertTrue(
            grammar.parse(
                '['
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,'
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,'
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,'
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,'
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,'
                'nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil'
                ']').is_valid)
        with self.assertRaises(MaxRecursionError):
            grammar.parse(
                '[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[hi'
                ']]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]'
            )


if __name__ == '__main__':
    unittest.main()



'''pyleri.List Class.

The grammar must find given elements separated by a delimiter. When no
delimiter is given we split using a comma. All arguments:

    element: List searches for this element
    delimiter: Elements must be separated by this delimiter. (default: ',')
    mi: Minimal elements the parser will search for. (default: 0)
    ma: Maximum elements the parser will search for. (default: None, unlimited)
    opt: When set to True the list may end with a delimiter. (default: False)

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
'''
from .elements import NamedElement, c_export, go_export, java_export


class List(NamedElement):

    __slots__ = ('_element', '_delimiter', '_min', '_max', '_opt')

    def __init__(self, element, delimiter=',', mi=0, ma=None, opt=False):
        self._element, self._delimiter = \
            map(self._validate_element, (element, delimiter))

        if not isinstance(mi, int) or mi < 0:
            raise TypeError('List(): "mi" must be an integer value larger '
                            'than or equal to 0, got: {}'.format(mi))
        self._min = mi

        if not isinstance(ma, (int, type(None))) or \
                (ma is not None and ma < mi):
            raise TypeError('List(): "ma" must be an integer or None value. '
                            'If an integer is specified it must be larger '
                            'than or equal to "mi"')
        self._max = ma

        # When opt is True the list may end with a delimiter
        self._opt = bool(opt)

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def opt_closing(self):
        return self._opt

    @property
    def _elements(self):
        yield self._element
        yield self._delimiter

    def _get_node_result(self, root, tree, rule, s, node):
        pos, i, j = node.start, 0, 0
        while True:
            is_valid, p = root._walk(
                self._element,
                pos,
                node.children,
                rule,
                i < self._min)
            if not is_valid:
                break
            i += 1
            pos = p
            if i == self._max and not self._opt:
                break

            is_valid, p = root._walk(
                self._delimiter,
                pos,
                node.children,
                rule,
                i < self._min)
            if not is_valid:
                break
            j += 1
            pos = p
            if j == self._max:
                break

        is_valid = not (i < self._min or (
            not self._opt and i and i == j))

        if is_valid:
            root._append_tree(tree, node, pos)

        return is_valid, pos

    def _run_export_js(self, js_indent, indent, classes, cname):
        return 'List({}, {}, {}, {}, {})'.format(
            self._element._export_js(js_indent, indent, classes, cname),
            self._delimiter._export_js(js_indent, indent, classes, cname),
            self._min,
            self._max or 'undefined',
            'true' if self._opt else 'false')

    def _run_export_py(self, py_indent, indent, classes):
        return 'List({}, {}, {}, {}, {})'.format(
            self._element._export_py(py_indent, indent, classes),
            self._delimiter._export_py(py_indent, indent, classes),
            self._min,
            self._max or 'None',
            'True' if self._opt else 'False')

    @c_export
    def _run_export_c(self, c_indent, indent, enums, gid):
        return 'cleri_list({}, {}, {}, {}, {}, {})'.format(
            gid,
            self._element._export_c(c_indent, indent, enums),
            self._delimiter._export_c(c_indent, indent, enums),
            self._min,
            self._max or '0',
            '1' if self._opt else '0')

    @go_export
    def _run_export_go(self, go_indent, indent, enums, gid):
        return 'goleri.NewList({}, {}, {}, {}, {}, {})'.format(
            gid,
            self._element._export_go(go_indent, indent, enums),
            self._delimiter._export_go(go_indent, indent, enums),
            self._min,
            self._max or '0',
            'true' if self._opt else 'false')

    @java_export
    def _run_export_java(self, java_indent, indent, enums, classes, gid):
        return 'new List({}{}, {}, {}, {}, {})'.format(
            '' if gid is None else 'Ids.{}, '.format(gid),
            self._element._export_java(java_indent, indent, enums, classes),
            self._delimiter._export_java(java_indent, indent, enums, classes),
            self._min,
            self._max or 'null',
            'true' if self._opt else 'false')



import unittest
import sys
import os
if os.environ.get('USELIB') != '1':
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from pyleri import (
    Grammar,
    Keyword,
    KeywordError,
    MissingStartError,
    NameAssignedError,
    Optional,
    ReKeywordsChangedError,
    UnusedElementError,
    create_grammar,
)  # nopep8
from examples import JsonGrammar  # nopep8


class TestCompile(unittest.TestCase):

    def test_compile_error(self):
        with self.assertRaises(ReKeywordsChangedError):
            class _G1(Grammar):
                k_test = Keyword('test')
                RE_KEYWORDS = '^[a-z]+'

        with self.assertRaises(KeywordError):
            class _G2(Grammar):
                k_test = Keyword('test-nomatch')

        with self.assertRaises(KeywordError):
            # test 'deep' keyword checks.
            class _G3(Grammar):
                START = Optional(Keyword('test-nomatch'))

        with self.assertRaises(NameAssignedError):
            class _G4(Grammar):
                k_test = Keyword('test')
                k_test2 = k_test

        with self.assertRaises(NameAssignedError):
            class _G5(Grammar):
                k_test = Keyword('test')
                k_Test = Keyword('duplicate')

        with self.assertRaises(MissingStartError):
            class _G6(Grammar):
                k_test = Keyword('test')

        with self.assertRaises(UnusedElementError):
            class _G7(Grammar):
                k_test = Keyword('test')
                START = Keyword('bla')

    def test_compile(self):
        class _G1(Grammar):
            k_hi = Keyword('hi')
            START = Optional(k_hi)

        self.assertEqual(_G1.k_hi.name, 'k_hi')


if __name__ == '__main__':
    unittest.main()
